'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import statsmodels.api as sm
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
class PValueSelector:
    def __init__(self, significance_level : float = 0.05) -> None:
        '''
            Setting up the algorithm
        :param significance_level: float, default = 0.05
            Used to compare the p-values with to find the features that must be removed from model
        '''
        self.significance_level = significance_level
        self.iter = 0

    def bin_to_cols(self, feature_state : list, X_column : list) -> list:
        '''
            Converting the binary state list into names of features that are included in model
        :param feature_state: list
            A list with zeros and ones used by model to understand what values should be pick by the model
            (when value = 1) and what it shouldn't pick (when value = 0)
        :param X_column: list
            The list with features of the dataframe before the algorithm is applied
        :return: list
            The list of features that where picked by the algorithm
        '''
        included_columns = [X_column[i] for i in range(len(feature_state)) if feature_state[i]==1]
        return included_columns

    def select(self, dataframe : 'pd.DataFrame', y_column : str) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied
        :param y_column: string
            The column name of the value that we what to predict
        :return: list
            The list of features that are selected by the algorithm as the best one
        '''
        # Defining the empty dictionary with the p-value history
        self.history = {}

        # Defining the list with names of columns except the predicted one
        X_columns = [col for col in dataframe.columns if col != y_column]
        for col in X_columns:
            self.history[col] = []

        # Defining the feature states
        feature_state = list(np.ones(len(X_columns)))
        while True:
            self.iter +=1

            # Extracting the selected columns
            X_cols = self.bin_to_cols(feature_state, X_columns)
            X = dataframe[X_cols].values
            y = dataframe[y_column].values

            # Fitting the OLS model from statsmodels
            regressor_OLS = sm.OLS(y, X).fit()

            # Getting the p-values from OLS model
            p_values = list(regressor_OLS.pvalues)
            p_values_index = 0

            # Adding p-values to history dictionary
            for col in X_columns:
                if col in X_cols:
                    self.history[col].append(float(p_values[p_values_index]))
                    p_values_index+=1
                else:
                    self.history[col].append(-1)

            # Choosing the max value of p-value
            max_PValue = max(p_values)

            # Erasing the column with the p-value equal with the max value of the p-value, if the max value is
            # higher than significance level
            if max_PValue > self.significance_level:
                for j in range(len(X_cols)):
                    if regressor_OLS.pvalues[j].astype(float) == max_PValue:
                        feature_state[X_columns.index(X_cols[j])] = 0
            else:
                break

        # Returning the chose columns.
        self.selected_cols = self.bin_to_cols(feature_state, X_columns)
        return self.selected_cols

    def plot_process(self, title : str = "P-Value Plot", save : bool = False, file_path : str = None) -> None:
        '''
            Ploting the process of finding the best features
        :param title string
            The title of the plot
        :param save: boolean, default = False
            If the this parameter is set to False that the model will not save the model
            If it is set to True the plot will be saved using :param file_path
        :param file_path: string, default = None
            The file path where the plot will be saved
            If the :param save is set to False the it is not used
        '''
        for col in self.history:
            line_style = lambda col: '-' if col in self.selected_cols else '--'
            plt.plot(range(self.iter), self.history[col], line_style(col), label=col)
        plt.hlines(self.significance_level, xmin=0, xmax=self.iter, colors='k')
        plt.xlabel("Iterations")
        plt.ylabel("p-values")
        plt.legend(ncol=3)
        plt.title(title)
        if save:
            plt.savefig(file_path)
        plt.show()


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
import pandas as pd
import math

def memoize(fun: 'function') -> 'function':
    '''
        The cache function.
    :param fun: function
        Function on which the decorator is applyed.
    :return: function
    '''
    results = dict()

    # Defining the internal helper function
    def helper(p: float):
        if p not in results:
            results[str(p)] = fun(p)
        return results[str(p)]

    return helper


@memoize
def info_theory(p: float) -> float:
    '''
        The information theory function.
    :param p: float
        The probability passed to function,
    :return: float
        The entropy.
    '''
    if p == 0:
        return 0
    else:
        return -p * math.log2(p)


class M3USelector:
    def __init__(self, num_features : int = 5, n_bins : int = 20) -> None:
        '''
            The Selector's constructor.
        :param num_features: int, default = 5
            The number of feature selected.
        :param n_bins: int, default = 20
            The number of bins to apply on the continous feature.
        '''
        self.__num_features = num_features
        self.__n_bins = n_bins

    def select(self, dataframe : pd.DataFrame, target : str) -> list:
        '''
            The selector function
        :param dataframe: pandas DataFrame
            The pandas DataFrame on which the function is applied.
        :param target: str
            The target column name.
        :return:
        '''
        self.selected_features = []
        self.X_columns = [col for col in list(dataframe.columns) if col != target]
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        for col in list(dataframe.columns):
            # Ignoring target column
            if col == target:
                continue

            # Checking if column is numeric
            if dataframe[col].dtype in numerics:
                # If column is numeric type we will apply bins split, else bins will be the values itself
                dataframe[col] = pd.qcut(dataframe[col], self.__n_bins, duplicates='drop', labels=False)
        for i in range(self.__num_features):
            # Defining the Minimal Mean Uncertainty Dictionary
            M2U = dict()

            # For every column wit already selected columns we calculates the Uncertainty
            for col in self.X_columns:
                if col in self.selected_features:
                    continue
                U = dict()
                MU = dict()
                temp = self.selected_features + [col]

                # Calculating the uncertainty for every column
                for temp_col in temp:
                    U[temp_col] = dict()
                    for q in dataframe[temp_col].unique():
                        U[temp_col][q] = sum(info_theory(p) for p in [len(dataframe[(dataframe[temp_col] == q) & (dataframe[target].isin(cls))]) / len(dataframe)
                                                                           for cls in [[dataframe[target].mode()], list(dataframe[dataframe[target] != dataframe[target].mode()[0]][target])]])
                for temp_col in temp:
                    MU[temp_col] = min(U[temp_col].values())
                M2U[col] = sum(MU.values()) / len(MU)
            selected_col = M2U.keys()

            # Selecting the column with the minimal mean minimal uncertainty
            for col in M2U:
                if M2U[col] == min(M2U.values()):
                    selected_col = col
            self.selected_features.append(selected_col)
        return self.selected_features


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
from functools import reduce
from operator import and_, or_
import pandas as pd

class MixerSelector:
    def __init__(self, selectors : list, strategy : str = 'intersection') -> None:
        '''
            Setting up the algorithm.
        :param selectors: list
            The list of initialized selectors.
        :param strategy: str, default = 'intersection'
            If set to 'union' the selector will return union of selected columns returned by selectors
            If set to 'intersection' the selector will return the intersection of columns returned by selectors
        '''
        self.__selectors = selectors
        self.__strategy = strategy

    def select(self, df : pd.DataFrame, target : str) -> list:
        '''
            Selecting the most important columns.
        :param df: pd.DataFrame
            The pandas DataFrame on which we want to apply feature selection
        :param target: str
            The column name of the value that we what to predict. Should be binary.
        :return: list
            The list of selected columns.
        '''
        # Defining the list where sets of columns for every selector will be stored
        selected_columns = []
        # Filling up the list with the columns selected by every selector
        for selector in self.__selectors:
            selected_columns.append(set(selector.select(df, target)))

        # Applying the chose strategy
        if self.__strategy == 'union':
            # Returning the union of the selected columns
            return list(reduce(or_, selected_columns))
        elif self.__strategy == 'intersection':
            # Returning the intersection of the selected columns
            return list(reduce(and_, selected_columns))
        else:
            # Raising an error if the strategy isn't the union or intersection
            raise ValueError(f"No such strategy as {self.__strategy}!")



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import numpy as np
import math

class ShannonSelector:
    def __init__(self, select_strategy : str = 'mean', nan_acceptance_level : float = 0.5) -> None:
        '''
            Setting up the algorithm
        :param select_strategy: str, default = 'mean'
            It defines the metric used to select the columns:
                'mean' : Use the mean value of information gain to select features
                'median' : Use the median value of information gain to select features
        :param nan_acceptance_level: float, default = 0.5
            The minimal quantity of NaN value in the column for feature to be accepted
        '''
        if select_strategy == 'mean':
            self.__select_strategy = np.mean
        elif select_strategy == 'median':
            self.__select_strategy = np.median
        else:
            raise ValueError("ShannonSelector doesn't accepts such a strategy. Choose 'mean' of 'median'")
        self.__nan_acceptance_level = nan_acceptance_level

    def __entropy(self, *classes_prob : list) -> float:
        '''
            This function calculates the entropy
        :param classes_prob: tuple
            The table with the normalized frequency of classes
        :return: float
            The entropy of sample
        '''
        entropy = 0
        for cls_prob in classes_prob:
            entropy -= cls_prob * math.log2(cls_prob)
        return entropy

    def select(self, dataframe : 'pd.DataFrame', y_column : str) -> list:
        '''
            This function implements the Shannon selection on the data frame
        :param dataframe: pandas.DataFrame
            The pandas DataFrame on which the algorithm must be applied
        :param y_column: str
            The string name of the target column
        :return: list
            The list of column names that where selected
        '''
        # Creating an empty dictionary to store the information gain for every column
        self.information_gains_ = dict()

        # The list of all column names except the target column
        X_columns = [col for col in dataframe.columns if col != y_column]

        # The number of sample for every class in the data set
        target_dict = dict(dataframe[y_column].value_counts())

        # The tuple with the names of every class name in the data set
        unique_classes_names = tuple(dataframe[y_column].unique())

        # The tuple with te normalised frequency of the target classed for the whole data set
        classes_prob = tuple(target_dict[unique_classes_names[i]]/len(dataframe) for i in range(len(unique_classes_names)))

        # The entropy of the whole data set based on the target clasees
        self.general_entropy_ = self.__entropy(*classes_prob)

        # Iterating throw dataframe columns
        for column in X_columns:
            # The dictionary with the classes count for the NaN values in the column
            nan_class_prob = dict(dataframe[dataframe[column].isna()][y_column].value_counts())

            # Skipping the columns with NaN values only in one class
            if len(nan_class_prob) < len(unique_classes_names):
                continue

            # Calculating the entropy for the samples with NaN values
            entropy_nan = self.__entropy(*tuple(nan_class_prob[class_name] / dataframe[column].isna().sum()
                                         for class_name in unique_classes_names))

            # Skipping the columns with NaN values in no class
            no_nan_class_prob = dict(dataframe[dataframe[column].notna()][y_column].value_counts())
            if len(no_nan_class_prob) < len(unique_classes_names):
                continue

            # Calculating the entropy for the samples without NaN values
            entropy_no_nan = self.__entropy(*tuple(no_nan_class_prob[class_name] / dataframe[column].notna().sum()
                                            for class_name in unique_classes_names))

            # Calculating the information gain for this column
            self.information_gains_[column] = self.general_entropy_ - entropy_nan * (
                        dataframe[column].isna().sum() / len(dataframe)) - entropy_no_nan * (dataframe[column].notna().sum() / len(dataframe))
        # Defining the columns where we will store all the selected columns
        saved_column = []

        for column in X_columns:
            # Selecting the columns with the number of NaN values smaller then __nan_acceptance_level or with the information gain smaller that mean or median of the information gains
            if dataframe[column].isna().sum() < self.__nan_acceptance_level * len(dataframe) or column in self.information_gains_ and self.information_gains_[column] < self.__select_strategy(list(self.information_gains_.values())):
                saved_column.append(column)
        return saved_column



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
@Author - Stojoc Vladimir - stojoc.vladimir@gmail.com
'''
import pandas as pd
from .errors import NotBetweenZeroAndOneError
from sklearn.decomposition import PCA

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class PCAReducer():
    def __init__(self, min_corr : float = 0.5, max_corr : float = 0.8, correlation_type : str = "pearson"):
        '''
            Setting up the algorithm
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with y_column
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with y_column
        :param correlation_type: str, default = "pearson" 
        '''
        try:
            if min_corr < 0 or max_corr > 1:
                raise NotBetweenZeroAndOneError
        except NotBetweenZeroAndOneError:
            print("Min or Max Correlations are not seted between 0 and 1!")
            quit()
        finally:
            self.max_corr = max_corr
            self.min_corr = min_corr
            self.correlation_type = correlation_type

    def index_to_cols(self, index_list):
        '''
            Converting the list of indexes in a list of features that will be picked by the model
        :param index_list: list
            A list of Indexes that should be converted into a list of feature names
        :return: list
            A list with feature names
        '''
        return [self.X_columns[i] for i in index_list]
        

    def reduce(self,dataframe : pd.DataFrame, y_column : str):
        '''
            Reducing the dimensionality of the data
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The column name of the value that we have to predict
        '''
        #Setting variables
        self.dataframe = dataframe
        self.X_columns = [col for col in dataframe.columns if col != y_column]
        correlated_indexes = []
        #Getting index of label column
        y_column_index = list(dataframe.columns).index(y_column)
        #Creating correlation table
        self.corr_table = dataframe.corr()
        corr_matrix = self.corr_table.values
        #Choosing most important columns
        for i in range(len(corr_matrix[y_column_index])):
            if abs(corr_matrix[y_column_index][i]) > self.min_corr and abs(corr_matrix[y_column_index][i]) < self.max_corr:
                correlated_indexes.append(i)
        self.correlated_cols = self.index_to_cols(correlated_indexes)
        #Creating correlation table for selected columns
        correlated_features_df = dataframe[self.correlated_cols]
        correlated_featurs_matrix = correlated_features_df.corr()
        self.correlated_features = []
        #Selecting most important correlations between features and creting lists of correlations
        for i in range(len(correlated_featurs_matrix)):
            cols = []
            for j in range(len(correlated_featurs_matrix)):
                if abs(correlated_featurs_matrix[self.correlated_cols[i]][self.correlated_cols[j]]) > self.min_corr and abs(correlated_featurs_matrix[self.correlated_cols[i]][self.correlated_cols[j]]) < self.max_corr :  # I also select the col on diagonal because it is also a part of correlation and should be in my array 
                    cols.append(self.correlated_cols[j])
            if len(cols)!=0 :
                cols.append(self.correlated_cols[i])
                cols.sort()
                self.correlated_features.append(cols) 
        #Cleaning selected columns and ordering them by length
        self.correlated_features = [list(item) for item in set(tuple(row) for row in self.correlated_features)]  
        self.correlated_features = sorted(self.correlated_features, key=len, reverse=True)
        self.dictionary = {}
        self.correlated_columns = [] 
        #Creting main dictionary with PCA filters for different correlations
        for i in range(len(self.correlated_features)):
            flag = False
            key = tuple(self.correlated_features[i])
            for j in range(len(self.correlated_features[i])):
                if self.correlated_features[i][j] not in dataframe.columns:
                    flag = True
            if flag:
                continue
            #Creating algorithm for every correlation
            self.correlated_columns.append(self.correlated_features[i]) # addding realy changed columns
            correlated_features_PCA_df = dataframe[self.correlated_features[i]].values
            pca = PCA(n_components = 1)
            pca.fit(correlated_features_PCA_df)
            X_pca = pca.transform(correlated_features_PCA_df)
            #Creating new column and adding it to dataframe
            column_name = "|".join(key)
            dataframe[column_name] = X_pca
            self.dictionary[key] = pca
            #Deleting correlated columns
            for j in range(len(self.correlated_features[i])):
                dataframe = dataframe.drop(self.correlated_features[i][j],1)
        return dataframe
    
    def apply(self, dataframe : pd.DataFrame, y_column : str):
        '''
            Reducing the dimensionality of the data
            based on an already existed filter
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The column name of the value that we have to predict
        '''
        for i in range(len(self.correlated_columns)):
            correlated_features_PCA_df = dataframe[self.correlated_columns[i]].values
            column_name = "|".join(self.correlated_columns[i])
            pca = self.dictionary[column_name]
            X_pca = pca.transform(correlated_features_PCA_df)
            dataframe[column_name] = X_pca
            for j in range(len(self.correlated_features[i])):
                dataframe = dataframe.drop(self.correlated_columns[i][j],1)
        return dataframe


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
import pandas as pd
from sklearn.decomposition import PCA

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class PCAFilter():
    def __init__(self, n_components : int = None): 
        '''
            Setting the algorithm
        :param n_components: integer, by default = None
            Number of components to keep
        '''
        self.n_components = n_components

    def reduce(self,dataframe : pd.DataFrame, y_column : str):
        '''
            Creating filter to new data and reducing the dimensionality of the data
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The column name of the value that we have to predict
        '''
        #Splitting dataframe
        self.dataframe = dataframe
        self.y_column = y_column
        self.X_columns = [col for col in self.dataframe.columns if col != self.y_column]
        self.X = dataframe[self.X_columns].values
        self.y = self.dataframe[y_column].values
        #Creating filter
        self.pca = PCA(n_components = self.n_components)
        self.pca.fit(self.X)
        #Creating new data based on the filter
        X_pca = self.pca.transform(self.X)
        X_new = self.pca.inverse_transform(X_pca)
        X_new = pd.DataFrame(X_new)
        #Create and return new Dataframe
        X_new[y_column] = self.y
        return X_new
    
    def apply(self,dataframe : pd.DataFrame, y_column : str):
        '''
            Reducing the dimensionality of the data
            based on an already existed filter
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The column name of the value that we have to predict
        '''
        #Splitting data
        X_columns = [col for col in dataframe.columns if col != y_column]
        X = dataframe[X_columns].values
        y = dataframe[y_column].values
        #Applying filter to the new dataframe
        X_pca = self.pca.transform(X)
        X_new = self.pca.inverse_transform(X_pca)
        X_new = pd.DataFrame(X_new)
        #Create and return new Dataframe
        X_new[y_column] = y
        return X_new


from .ANOVASelector import ANOVASelector
from .ChiSquaredSelector import ChiSquaredSelector
from .KendallCorrelationSelector import KendallCorrelationSelector
from .LassoSelector import LassoSelector
from .PearsonCorrelationSelector import PearsonCorrelationSelector
from .PointBiserialSelector import PointBiserialCorrSelector
from .PValueSelector import PValueSelector
from .SpearmanCorrelationSelector import SpearmanCorrelationSelector
from .GeneticAlgorithmSelector import GeneticAlgorithmSelector
from .ShannonSelector import ShannonSelector
from .LDAReducer import LDAReducer
from .M3USelector import M3USelector
from .MixerSelector import MixerSelector
from .MUSESelector import MUSESelector
from .ElasticNet import ElasticNetSelector
from .PCAReducer import PCAReducer
from .PCAFilter import PCAFilter



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import pandas as pd
import math

class NotBinaryData(BaseException):
    pass

class MUSESelector():
    def __init__(self, num_features : int = 5, n_bins : int = 20, p : float = 0.2, T : float = 0.1) -> None:
        '''
            Setting up the algorithm.
        :param num_features: int, default = 5
            The number of features that will remain after the algorithm will be applied.
        :param n_bins: int, default = 20
            The number of bins in which the series will be split if it's continuous
        :param p: float, default = 0.2
            The minimal value for cumulated sum of probabilities of positive class frequency
        :param T: float, default = 0.1
            The minimal threshold of the class impurity for a bin to be passed to the selector
        '''
        self.__num_features = num_features
        self.__n_bins = n_bins
        self.__p = p
        self.__T = T

    def select(self, df : pd.DataFrame, target : str) -> list:
        '''
            Selecting the most important columns.
        :param df: pd.DataFrame
            The pandas DataFrame on which we want to apply feature selection
        :param target: str
            The column name of the value that we what to predict. Should be binary.
        :return: list
            The list of selected columns.
        '''
        # Checking if the target column is binary one
        if len(df[target].unique()) != 2:
            raise NotBinaryData(f"{target} column isn't a binary column")

        # Creating an empty dictionary to store metadata about features
        features_data = dict()
        self.X_columns = [col for col in df.columns if col != target]
        self.selected_features = []

        # Defining the pandas numeric types of pandas
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        for col in list(df.columns):
            # Ignoring target column
            if col == target:
                continue

            # Checking if column is numeric
            if df[col].dtype in numerics:
                # If column is numeric type we will apply bins split, else bins will be the values itself
                df[col] = pd.qcut(df[col], self.__n_bins, duplicates='drop', labels=False)

            # Defining the metadata about features
            labels = dict()
            labels['impurity'] = dict()
            labels['H'] = dict()
            labels['discarded'] = [].copy()

            # Calculating the metadata for every bin of the feature
            for q in df[col].unique():
                labels['impurity'][q] = min(prob / len(df) for prob in dict(df[df[col]==q][target].value_counts()).values())
                labels['H'][q] = sum(-p / len(df) * math.log2(p / len(df)) for p in dict(df[df[col]==q][target].value_counts()).values())
            features_data[col] = labels

        # Selecting the 'num_features' features
        for i in range(self.__num_features):
            # Defining the selecting criteria J
            J = dict()

            for feature in features_data.keys():
                # Skipping already selected features
                if feature in self.selected_features:
                    continue

                # Defining a empty list where selected bins will be stored
                selected_intervals = [].copy()
                H = features_data[feature]['H']

                # Selecting bins with the cumulated sum of p smaller than self.p
                p_sum = 0
                for key, value in {k: v for k, v in sorted(H.items(), key=lambda item: item[1])}.items():
                    if key in features_data[feature]['discarded']:
                        continue
                    selected_intervals.append(key)
                    p_sum += value
                    if p_sum > self.__p:
                        break

                # Calculating the J for feature
                J[feature] = sum(len(df[df[feature] == q])/len(df)*H[q] for q in selected_intervals)
            # Searching for the minimal value of J
            min_j_feature = list(J.keys())[0]
            for feature in J:
                if J[feature] < J[min_j_feature]:
                    min_j_feature = feature

            # Adding the feature with the minimal value of J to the selected_features
            self.selected_features.append(min_j_feature)

            # Discarding bins with the impurity smaller than self.T
            for feature in features_data:
                for i in range(len(features_data[feature]['impurity'])):
                    if features_data[feature]['impurity'][i] < self.__T:
                        features_data[feature]['discarded'].append(i)
        return self.selected_features
